@@ -23,13 +23,20 @@ import time
 
 import paramiko
 
+CONNECTED = 'connected'
+CONNECTING = 'connecting'
+CONSUMING = 'consuming'
+DEAD = 'dead'
+DISCONNECTED = 'disconnected'
+IDLE = 'idle'
 
-class GerritWatcher(object):
+
+class GerritWatcher(threading.Thread):
     log = logging.getLogger("gerrit.GerritWatcher")
 
     def __init__(
             self, gerrit, username=None, hostname=None, port=None,
-            keyfile=None):
+            keyfile=None, connection_attempts=-1, retry_delay=5):
         """Create a GerritWatcher.
 
         :param gerrit: A Gerrit instance to pass events to.
@@ -37,11 +44,16 @@ class GerritWatcher(object):
         All other parameters are optional and if not supplied are sourced from
         the gerrit instance.
         """
+        super(GerritWatcher, self).__init__()
+        assert retry_delay >= 0, "Retry delay must be >= 0"
         self.username = username or gerrit.username
         self.keyfile = keyfile or gerrit.keyfile
         self.hostname = hostname or gerrit.hostname
         self.port = port or gerrit.port
         self.gerrit = gerrit
+        self.connection_attempts = int(connection_attempts)
+        self.retry_delay = float(retry_delay)
+        self.state = IDLE
 
     def _read(self, fd):
         l = fd.readline()
@@ -62,32 +74,99 @@ class GerritWatcher(object):
                     else:
                         raise Exception("event on ssh connection")
 
-    def _run(self):
-        try:
+    def _connect(self):
+        """Attempts to connect and returns the connected client."""
+
+        def _make_client():
             client = paramiko.SSHClient()
             client.load_system_host_keys()
             client.set_missing_host_key_policy(paramiko.WarningPolicy())
-            client.connect(self.hostname,
-                           username=self.username,
-                           port=self.port,
-                           key_filename=self.keyfile)
+            return client
 
-            stdin, stdout, stderr = client.exec_command("gerrit stream-events")
+        def _attempt_gen(connection_attempts, retry_delay):
+            if connection_attempts <= 0:
+                attempt = 1
+                while True:
+                    yield (attempt, retry_delay)
+                    attempt += 1
+            else:
+                for attempt in range(1, connection_attempts + 1):
+                    if attempt < connection_attempts:
+                        yield (attempt, retry_delay)
+                    else:
+                        # No more attempts left after this one, (signal this by
+                        # not returning a valid retry_delay).
+                        yield (attempt, None)
 
-            self._listen(stdout, stderr)
+        for (attempt, retry_delay) in _attempt_gen(self.connection_attempts,
+                                                   self.retry_delay):
+            self.log.debug("Connection attempt %s to %s:%s (retry_delay=%s)",
+                           attempt, self.hostname, self.port, retry_delay)
+            client = None
+            try:
+                client = _make_client()
+                client.connect(self.hostname,
+                               username=self.username,
+                               port=self.port,
+                               key_filename=self.keyfile)
+                return client
+            except (IOError, paramiko.SSHException) as e:
+                self.log.exception("Exception connecting to %s:%s",
+                                   self.hostname, self.port)
+                if client:
+                    try:
+                        client.close()
+                    except (IOError, paramiko.SSHException):
+                        self.log.exception("Failure closing broken client")
+                if retry_delay is not None:
+                    if retry_delay > 0:
+                        self.log.info("Trying again in %s seconds",
+                                      retry_delay)
+                        time.sleep(retry_delay)
+                else:
+                    raise e
 
-            ret = stdout.channel.recv_exit_status()
-            self.log.debug("SSH exit status: %s" % ret)
+    def _consume(self, client):
+        """Consumes events using the given client."""
+        stdin, stdout, stderr = client.exec_command("gerrit stream-events")
 
-            if ret:
-                raise Exception("Gerrit error executing stream-events")
+        self.state = CONSUMING
+        self._listen(stdout, stderr)
+
+        ret = stdout.channel.recv_exit_status()
+        self.log.debug("SSH exit status: %s" % ret)
+
+        if ret:
+            raise Exception("Gerrit error executing stream-events:"
+                            " return code %s" % ret)
+
+    def _run(self):
+        self.state = CONNECTING
+        client = self._connect()
+        self.state = CONNECTED
+        try:
+            self._consume(client)
         except Exception:
-            self.log.exception("Exception on ssh event stream:")
-            time.sleep(5)
+            # NOTE(harlowja): allow consuming failures to *always* be retryable
+            self.log.exception("Exception consuming ssh event stream:")
+            if client:
+                try:
+                    client.close()
+                except (IOError, paramiko.SSHException):
+                    self.log.exception("Failure closing broken client")
+            self.state = DISCONNECTED
+            if self.retry_delay > 0:
+                self.log.info("Delaying consumption retry for %s seconds",
+                              self.retry_delay)
+                time.sleep(self.retry_delay)
 
     def run(self):
-        while True:
-            self._run()
+        try:
+            while True:
+                self.state = DISCONNECTED
+                self._run()
+        finally:
+            self.state = DEAD
 
 
 class Gerrit(object):
@@ -100,10 +179,14 @@ class Gerrit(object):
         self.keyfile = keyfile
         self.watcher_thread = None
         self.event_queue = None
+        self.installed_plugins = None
 
-    def startWatching(self):
+    def startWatching(self, connection_attempts=-1, retry_delay=5):
         self.event_queue = Queue.Queue()
-        self.watcher_thread = threading.Thread(target=GerritWatcher(self).run)
+        watcher = GerritWatcher(self,
+                                connection_attempts=connection_attempts,
+                                retry_delay=retry_delay)
+        self.watcher_thread = watcher
         self.watcher_thread.daemon = True
         self.watcher_thread.start()
 
@@ -119,14 +202,16 @@ class Gerrit(object):
             cmd = '%s --visible-to-all' % cmd
         if owner:
             cmd = '%s --owner %s' % (cmd, owner)
-        cmd = '%s %s' % (cmd, group)
+        cmd = '%s "%s"' % (cmd, group)
         out, err = self._ssh(cmd)
         return err
 
-    def createProject(self, project, require_change_id=True):
+    def createProject(self, project, require_change_id=True, empty_repo=False):
         cmd = 'gerrit create-project'
         if require_change_id:
             cmd = '%s --require-change-id' % cmd
+        if empty_repo:
+            cmd = '%s --empty-commit' % cmd
         cmd = '%s --name %s' % (cmd, project)
         out, err = self._ssh(cmd)
         return err
@@ -134,15 +219,39 @@ class Gerrit(object):
     def listProjects(self):
         cmd = 'gerrit ls-projects'
         out, err = self._ssh(cmd)
-        return out.split('\n')
+        return filter(None, out.split('\n'))
 
     def listGroups(self):
         cmd = 'gerrit ls-groups'
         out, err = self._ssh(cmd)
-        return out.split('\n')
+        return filter(None, out.split('\n'))
+
+    def listPlugins(self):
+        plugins = self.getPlugins()
+        plugin_names = plugins.keys()
+        return plugin_names
+
+    # get installed plugins info returned is (name, version, status, file)
+    def getPlugins(self):
+        # command only available on gerrit verion >= 2.5
+        cmd = 'gerrit plugin ls --format json'
+        out, err = self._ssh(cmd)
+        return json.loads(out)
+
+    def getVersion(self):
+        # command only available on gerrit verion >= 2.6
+        cmd = 'gerrit version'
+        out, err = self._ssh(cmd)
+        out = out.split(' ')[2]
+        return out.strip('\n')
 
     def replicate(self, project='--all'):
-        cmd = 'gerrit replicate %s' % project
+        cmd = 'replication start %s' % project
+        if self.installed_plugins is None:
+            try:
+                self.installed_plugins = self.listPlugins()
+            except Exception:
+                cmd = 'gerrit replicate %s' % project
         out, err = self._ssh(cmd)
         return out.split('\n')
 
